import base64
import datetime
import hashlib
import json
import logging
import pytz
import requests
import secrets

from decimal import Decimal
from flask import current_app as app
from flask_babel import lazy_gettext as _
from typing import List

from cryptoadvance.specter.wallet import Wallet

from .service import SwanService


logger = logging.getLogger(__name__)


# TODO: Update with prod values
client_id = "specter-dev"
client_secret = "BcetcVcmueWf5P3UPJnHhCBMQ49p38fhzYwM7t3DJGzsXSjm89dDR5URE46SY69j"
code_verifier = "64fRjTuy6SKqdC1wSoInUNxX65dQUhVVKTqZXuQ7dqw"
api_url = app.config.get("SWAN_API_URL")


class SwanApiException(Exception):
    pass


class SwanApiRefreshTokenException(SwanApiException):
    pass


def get_oauth2_start_url():
    """
    Set up the Swan API integration by requesting our initial access_token and
    refresh_token.
    """
    # Let's start the PKCE-flow
    global code_verifier

    if code_verifier is None:
        code_verifier = secrets.token_urlsafe(43)
    # see specification: https://datatracker.ietf.org/doc/html/rfc7636#section-4.2
    # and example impl: https://github.com/RomeoDespres/pkce/blob/master/pkce/__init__.py#L94-L96
    hashed = hashlib.sha256(code_verifier.encode("ascii")).digest()
    encoded = base64.urlsafe_b64encode(hashed)
    code_challenge = encoded.decode("ascii")[:-1]

    flow_url = f"{api_url}/oidc/auth?"
    query_params = [
        "client_id=specter-dev",
        "redirect_uri=http://localhost:25441/svc/swan/oauth2/callback",  # TODO: Will localhost work in all usage contexts (e.g. standalone app)?
        "response_type=code",
        "response_mode=query",
        f"code_challenge={code_challenge}",
        "code_challenge_method=S256",
        "state=kjkmdskdmsmmsmdslmdlsm",
        "scope=offline_access v1 write:vendor_wallet read:vendor_wallet write:automatic_withdrawal read:automatic_withdrawal",
        "prompt=consent",
    ]
    flow_url += "&".join(query_params)

    return flow_url


def get_access_token(code: str = None, code_verifier: str = None):
    """
    If code and code_verifier are specified, this is our initial request for an
    access_token and, more importantly, the refresh_token.

    If code is None, use the refresh_token to get a new short-lived access_token.

    If we don't have the refresh_token, raise SwanApiRefreshTokenException.
    """
    if code:
        # Requesting initial refresh_token and access_token
        payload = {
            "client_id": "specter-dev",
            "client_secret": client_secret,
            "code_verifier": code_verifier,
            "grant_type": "authorization_code",
            "code": code,
        }
        auth_header = None
    else:
        service_data = SwanService.get_current_user_service_data()
        if SwanService.is_access_token_valid():
            return service_data[SwanService.ACCESS_TOKEN]

        # Use the refresh_token to get a new access_token
        if SwanService.REFRESH_TOKEN not in service_data:
            raise SwanApiRefreshTokenException(
                "access_token is expired but we don't have a refresh_token"
            )

        payload = {
            "grant_type": "refresh_token",
            # "redirect_uri":   # Necessary?
            "refresh_token": service_data[SwanService.REFRESH_TOKEN],
            "scope": "offline_access v1 write:vendor_wallet read:vendor_wallet write:automatic_withdrawal read:automatic_withdrawal",
        }

        auth_hash = base64.b64encode(f"{client_id}:{client_secret}".encode()).decode()
        auth_header = {
            "Authorization": f"Basic {auth_hash}",
        }
        logger.debug(f"auth_hash: {auth_hash}")
        logger.debug("Using the refresh_token to request an access_token")
        logger.debug(f"payload: {json.dumps(payload, indent=4)}")

    logger.debug(payload)
    logger.debug(auth_header)

    response = requests.post(
        f"{api_url}/oidc/token",
        data=payload,
        headers=auth_header,
    )
    logger.debug(f"api_url: {api_url}")
    logger.debug(response)
    logger.debug(response.text)
    resp = json.loads(response.text)
    """
        {
            "access_token": "***************",
            "expires_in": 3600,
            "refresh_token": "***************",
            "scope": "offline_access v1 write:vendor_wallet read:vendor_wallet write:automatic_withdrawal read:automatic_withdrawal",
            "token_type": "Bearer"
        }
    """
    # TODO: Remove debugging
    logger.debug(json.dumps(resp, indent=4))
    if resp.get("access_token"):
        new_api_data = {
            SwanService.ACCESS_TOKEN: resp["access_token"],
            SwanService.ACCESS_TOKEN_EXPIRES: (
                datetime.datetime.now(tz=pytz.utc)
                + datetime.timedelta(seconds=resp["expires_in"])
            ).timestamp(),
        }
        if "refresh_token" in resp:
            new_api_data[SwanService.REFRESH_TOKEN] = resp["refresh_token"]

        SwanService.update_current_user_service_data(new_api_data)

        # TODO: Remove debugging
        logger.debug(
            f"service_data: {json.dumps(SwanService.get_current_user_service_data(), indent=4)}"
        )
        return resp["access_token"]
    else:
        logger.warning(response)
        raise SwanApiException(response.text)


def handle_oauth2_auth_callback(request):
    code = request.args.get("code")
    logger.debug(f"request.args: {request.args}")
    logger.debug(f"looks good, we got a code: {code}")
    get_access_token(code=code, code_verifier=code_verifier)


def authenticated_request(
    endpoint: str, method: str = "GET", json_payload: dict = {}
) -> dict:
    logger.debug(f"{method} endpoint: {endpoint}")

    access_token = get_access_token()

    auth_header = {
        "Authorization": f"Bearer {access_token}",
    }
    try:
        if method == "GET":
            response = requests.get(api_url + endpoint, headers=auth_header)
        elif method in ["POST", "PATCH", "PUT", "DELETE"]:
            response = requests.request(
                method=method,
                url=api_url + endpoint,
                headers=auth_header,
                json=json_payload,
            )
        if response.status_code != 200:
            raise SwanApiException(f"{response.status_code}: {response.text}")
        return response.json()
    except Exception as e:
        # TODO: tighten up expected Exceptions
        logger.exception(e)
        logger.error(
            f"endpoint: {endpoint} | method: {method} | payload: {json.dumps(json_payload, indent=4)}"
        )
        logger.error(f"{response.status_code}: {response.text}")
        raise e


def get_autowithdrawal_addresses(swan_wallet_id: str = None) -> dict:
    """
    {
        "entity": "wallet",
        "item": {
            "id": "c47e1e83-90a0-45da-ae25-6a0d324b9f29",
            "isConfirmed": false,
            "displayName": "Specter autowithdrawal to SeedSigner demo",
            "metadata": {
                "oidc": {
                    "clientId": "specter-dev"
                },
                "specter_wallet_alias": "seedsigner_demo"
            },
            "btcAddresses": []
        }
    }
    """
    if not swan_wallet_id:
        swan_wallet_id = SwanService.get_current_user_service_data().get(
            SwanService.SWAN_WALLET_ID
        )

    resp = authenticated_request(
        endpoint=f"/apps/v20210824/wallets/{swan_wallet_id}?full=true",
        method="GET",
    )

    logger.debug(json.dumps(resp, indent=4))
    return resp


def update_autowithdrawal_addresses(
    specter_wallet_name: str, specter_wallet_alias: str, addresses: List[str]
) -> dict:
    """
    * If SWAN_WALLET_ID is known, any existing unused addresses are cleared.
    * If there is no known SWAN_WALLET_ID, we `POST` to create an initial Swan wallet and store the resulting SWAN_WALLET_ID.
    * Sends the list of new addresses for SWAN_WALLET_ID.
    """
    swan_wallet_id = SwanService.get_current_user_service_data().get(
        SwanService.SWAN_WALLET_ID
    )

    if swan_wallet_id:
        # We already have a Swan walletId. DELETE the existing unused addresses...
        delete_autowithdrawal_addresses(swan_wallet_id)

        # ...and then append the new ones.
        endpoint = f"/apps/v20210824/wallets/{swan_wallet_id}/addresses"
        method = "PATCH"
    else:
        # We don't yet have a Swan walletId. POST to create one.
        endpoint = "/apps/v20210824/wallets"
        method = "POST"

    resp = authenticated_request(
        endpoint=endpoint,
        method=method,
        json_payload={
            "btcAddresses": [{"address": addr} for addr in addresses],
            "displayName": str(
                _('Specter Desktop "{}"').format(specter_wallet_name)
            ),  # Can't pass a LazyString into json
            "metadata": {
                "specter_wallet_alias": specter_wallet_alias,
            },
        },
    )

    """
    Response should include wallet ("item") details:
        {
            "entity": "wallet",
            "item": {
                "id": "c47e1e83-90a0-45da-ae25-6a0d324b9f29",
                "isConfirmed": false,
                "displayName": "Specter autowithdrawal to SeedSigner demo",
                "metadata": {
                    "oidc": {
                        "clientId": "specter-dev"
                    },
                    "specter_wallet_alias": "seedsigner_demo"
                },
                "btcAddresses": []
            }
        }
    """
    logger.debug(json.dumps(resp, indent=4))

    if "item" in resp and "id" in resp["item"]:
        if resp["item"]["id"] != swan_wallet_id:
            swan_wallet_id = resp["item"]["id"]

            # Save the swan_wallet_id in the user's persistent Service settings
            logger.debug(f"Updating the Swan wallet id to {swan_wallet_id}")
            SwanService.update_current_user_service_data(
                {SwanService.SWAN_WALLET_ID: swan_wallet_id}
            )
    else:
        raise SwanApiException(
            f"No 'id' returned for the new/updated wallet: {json.dumps(resp, indent=4)}"
        )


def delete_autowithdrawal_addresses(swan_wallet_id: str):
    """
    Deletes all unused autowithdrawal addresses from the specified SWAN_WALLET_ID
    """
    resp = authenticated_request(
        endpoint=f"/apps/v20210824/wallets/{swan_wallet_id}/addresses",
        method="DELETE",
    )
    logger.debug(json.dumps(resp, indent=4))
    return resp


def get_autowithdrawal_info() -> dict:
    """
    See note in set_autowithdrawal. This returns all autowithdrawal objs from the Swan
    side.
    """
    resp = authenticated_request(
        endpoint="/apps/v20210824/automatic-withdrawal",
        method="GET",
    )
    logger.debug(json.dumps(resp, indent=4))
    return resp


def set_autowithdrawal(btc_threshold: Decimal) -> dict:
    """
    0 == Weekly; other float values = BTC threshold

    The Swan api generates a new autowithdrawal id each time but there is no support to
    update an existing autowithdrawal, other than activating or deactivating it.

    New autowithdrawals are initialized as `isActive: false` and require the user to
    complete a Swan-side email verification step.

    We save the resulting autowithdrawal_id even though it isn't clear at the moment if
    it's desirable to ever call the `deactivate/` or `activate/` endpoints.
    """
    service_data = SwanService.get_current_user_service_data()
    swan_wallet_id = service_data.get(SwanService.SWAN_WALLET_ID)

    endpoint = "/apps/v20210824/automatic-withdrawal"
    method = "POST"
    resp = authenticated_request(
        endpoint=endpoint,
        method=method,
        json_payload={
            "walletId": swan_wallet_id,
            "minBtcThreshold": btc_threshold,
        },
    )
    logger.debug(json.dumps(resp, indent=4))

    """
        {
            "entity": "automaticWithdrawal",
            "item": {
                "id": "******************",
                "minBtcThreshold": "0.01",
                "isActive": false,
                "isCanceled": false,
                "createdAt": "2022-01-07T02:14:56.070Z",
                "walletId": "******************",
                "walletAddressId": null
            }
        }
    """
    if "item" in resp and "id" in resp["item"]:
        autowithdrawal_id = resp["item"]["id"]
        if autowithdrawal_id != service_data.get(SwanService.AUTOWITHDRAWAL_ID):
            SwanService.update_current_user_service_data(
                {
                    SwanService.AUTOWITHDRAWAL_ID: autowithdrawal_id,
                }
            )
    else:
        raise SwanApiException(
            f"No 'id' returned for the new/updated autowithdrawal: {json.dumps(resp, indent=4)}"
        )

    return resp


def activate_autowithdrawal() -> dict:
    """
    Activates the autowithdrawal specified in SwanService.AUTOWITHDRAWAL_ID.

    If the automatic withdrawal was just created, this will generate a 400 error:
    "Cannot activate an automatic withdrawal before withdrawal address is confirmed".

    The user must first confirm the first withdrawal addr via Swan-side email flow.
    After they confirm, the autowithdrawal should then return `isActive: true`.

    NOT CURRENTLY USED; remove if we don't ever enable disable/activate flows.
    """
    service_data = SwanService.get_current_user_service_data()
    autowithdrawal_id = service_data.get(SwanService.AUTOWITHDRAWAL_ID)
    if not autowithdrawal_id:
        raise SwanApiException(
            f"AUTOWITHDRAWAL_ID ({SwanService.AUTOWITHDRAWAL_ID}) not found in service data"
        )

    endpoint = f"/apps/v20210824/automatic-withdrawal/{autowithdrawal_id}/activate"
    method = "POST"
    resp = authenticated_request(
        endpoint=endpoint,
        method=method,
    )
    logger.debug(json.dumps(resp, indent=4))

    """
        {
            "entity": "automaticWithdrawal",
            "item": {
                "id": "******************",
                "minBtcThreshold": "0.01",
                "isActive": true,
                "isCanceled": false,
                "createdAt": "2022-01-07T02:14:56.070Z",
                "walletId": "******************",
                "walletAddressId": null
            }
        }
    """
    if "item" in resp and "id" in resp["item"]:
        if resp["item"]["id"] != service_data[SwanService.AUTOWITHDRAWAL_ID]:
            autowithdrawal_id = resp["item"]["id"]
    else:
        raise SwanApiException(
            f"No 'id' returned for the new/updated autowithdrawal: {json.dumps(resp, indent=4)}"
        )

    return resp


def get_wallet_details(swan_wallet_id: str) -> dict:
    """
    {
        "entity": "wallet",
        "item": {
            "id": "********",
            "isConfirmed": false,
            "displayName": "Specter autowithdrawal to SeedSigner demo",
            "metadata": {
                "oidc": {
                    "clientId": "specter-dev"
                },
                "specter_wallet_alias": "seedsigner_demo"
            },
            "btcAddresses": []
        }
    }
    """
    resp = authenticated_request(
        endpoint=f"/apps/v20210824/wallets/{swan_wallet_id}",
        method="GET",
    )

    logger.debug(json.dumps(resp, indent=4))
    return resp


def get_wallets() -> dict:
    """
    Return all Swan wallet entries. Should only be one per Specter-Swan user combo (but can be more due
    to testing/debugging, calling `/wallets` POST more than once, etc.)
    """
    resp = authenticated_request(
        endpoint=f"/apps/v20210824/wallets",
        method="GET",
    )
    """
        {
            "entity": "wallet",
            "list": [
                {
                    "id": "**********",
                    "walletAddressId": "**********",
                    "btcAddress": "bc1q**********",
                    "isConfirmed": false,
                    "displayName": "Specter Desktop \"SeedSigner demo\"",
                    "metadata": {
                        "oidc": {
                            "clientId": "specter-dev"
                        },
                        "specter_wallet_alias": "seedsigner_demo"
                    }
                },
                {
                    "id": "**********",
                    "walletAddressId": "**********",
                    "btcAddress": "bc1q**********",
                    "isConfirmed": false,
                    "displayName": "Specter Desktop \"DCA Corn\"",
                    "metadata": {
                        "oidc": {
                            "clientId": "specter-dev"
                        },
                        "specter_wallet_alias": "dca_corn_2"
                    }
                },
                ...,
            ]
        }
    """
    logger.debug(json.dumps(resp, indent=4))

    return resp



import logging
import os
from importlib import import_module

from flask import current_app as app
from flask.blueprints import Blueprint
from cryptoadvance.specter.wallet import Wallet
from flask_babel import lazy_gettext as _
from typing import List

from .service_encrypted_storage import ServiceEncryptedStorageManager
from .service_annotations_storage import ServiceAnnotationsStorage

from cryptoadvance.specter.addresslist import Address


logger = logging.getLogger(__name__)


devstatus_alpha = "alpha"
devstatus_beta = "beta"
devstatus_prod = "prod"


class Service:
    """A base class for Services"""

    # These should be overrided in implementation classes
    id = None
    name = None
    icon = None
    logo = None
    desc = None  # TODO: rename to "description" to be explicit
    has_blueprint = True  # the default
    devstatus = devstatus_alpha

    def __init__(self, active, specter):
        if not hasattr(self, "id"):
            raise Exception(f"Service {self.__class__} needs ID")
        if not hasattr(self, "name"):
            raise Exception(f"Service {self.__class__} needs name")
        self.active = active
        self.specter = specter
        if self.has_blueprint:
            self.__class__.blueprint = Blueprint(
                f"{self.id}_endpoint",
                f"cryptoadvance.specter.services.{self.id}.service",  # To Do: move to subfolder
                template_folder="templates",
                static_folder="static",
            )

            def inject_stuff():
                """Can be used in all jinja2 templates"""
                return dict(specter=app.specter, service=self)

            self.__class__.blueprint.context_processor(inject_stuff)
            # Import the controller for this service
            import_module(f"cryptoadvance.specter.services.{self.id}.controller")
            app.register_blueprint(
                self.__class__.blueprint, url_prefix=f"/svc/{self.id}"
            )

    @classmethod
    def set_current_user_service_data(cls, service_data: dict):
        ServiceEncryptedStorageManager.get_instance().set_current_user_service_data(
            service_id=cls.id, service_data=service_data
        )

    @classmethod
    def update_current_user_service_data(cls, service_data: dict):
        ServiceEncryptedStorageManager.get_instance().update_current_user_service_data(
            service_id=cls.id, service_data=service_data
        )

    @classmethod
    def get_current_user_service_data(cls) -> dict:
        return (
            ServiceEncryptedStorageManager.get_instance().get_current_user_service_data(
                service_id=cls.id
            )
        )

    @classmethod
    def get_blueprint_name(cls):
        return f"{cls.id}_endpoint"

    @classmethod
    def default_address_label(cls):
        # Have to str() it; can't pass a LazyString to json serializer
        return str(_("Reserved for {}").format(cls.name))

    @classmethod
    def reserve_address(cls, wallet: Wallet, address: str, label: str = None):
        # Mark an Address in a persistent way as being reserved by a Service
        if not label:
            label = cls.default_address_label()
        wallet.associate_address_with_service(
            address=address, service_id=cls.id, label=label
        )

    @classmethod
    def reserve_addresses(
        cls,
        wallet: Wallet,
        label: str = None,
        num_addresses: int = 10,
        annotations: dict = None,
    ) -> List[str]:
        """
        Reserve n unused addresses but leave a gap between each one so that this reserved
        range never causes an address gap in the wallet (e.g. if you reserve ten in a
        row it's possible that some wallet software will miss a new tx on the 11th
        address).

        If `label` is not provided, we use Service.default_address_label().

        Optional `annotations` data can be attached to each Address being reserved.
        """
        # Track Service-related addresses in ServiceAnnotationsStorage
        if annotations:
            annotations_storage = ServiceAnnotationsStorage(
                service_id=cls.id, wallet=wallet
            )

        addresses = []
        start_index = wallet.address_index + 1
        for i in range(start_index, start_index + (2 * num_addresses), 2):
            address = wallet.get_address(i)

            # Mark an Address in a persistent way as being reserved by a Service
            cls.reserve_address(wallet=wallet, address=address)

            addresses.append(address)

            if annotations:
                annotations_storage.set_addr_annotations(
                    addr=address, annotations=annotations, autosave=False
                )
        if annotations:
            annotations_storage.save()

        return addresses

    @classmethod
    def unreserve_addresses(cls, wallet: Wallet):
        """
        Clear out Services-related data from any unused Addresses, but leave already-used
        Addresses as-is.
        """
        annotations_storage = ServiceAnnotationsStorage(
            service_id=cls.id, wallet=wallet
        )
        addrs = wallet.get_associated_addresses(service_id=cls.id, unused_only=True)
        for addr_obj in addrs:
            wallet.deassociate_address(addr_obj["address"])
            annotations_storage.remove_addr_annotations(
                addr_obj.address, autosave=False
            )
        annotations_storage.save()

    # def is_active(self):
    #     return self.active

    # def set_active(self, value):
    #     self.active = value

    @classmethod
    def update(self):
        """
        Called by backend periodic process to keep Service in sync with any remote
        data (e.g. fetching the latest data from an external API).
        """
        logger.info(f"update() not implemented / not necessary for Service {self.id}")

    """ ***********************************************************************
                                    Update hooks
    *********************************************************************** """

    @classmethod
    def on_user_login(cls):
        pass



import json
import logging
import os

from flask import current_app as app, url_for
from flask.blueprints import Blueprint
from importlib import import_module
from inspect import isclass
from pathlib import Path
from pkgutil import iter_modules
from typing import Dict, List
from cryptoadvance.specter.user import User

from cryptoadvance.specter.managers.singleton import ConfigurableSingletonException
from ..util.reflection import get_subclasses_for_class

from .service import Service
from .service_encrypted_storage import ServiceEncryptedStorageManager

logger = logging.getLogger(__name__)


class ServiceManager:
    """Loads support for all Services it auto-discovers."""

    def __init__(self, specter, devstatus_threshold):
        self.specter = specter
        self.devstatus_threshold = devstatus_threshold

        # Each Service class is stored here, keyed on its Service.id str
        self._services: Dict[str, Service] = {}
        logger.info("----> starting service discovery <----")
        for clazz in get_subclasses_for_class(Service):
            compare_map = {"alpha": 1, "beta": 2, "prod": 3}
            if compare_map[self.devstatus_threshold] <= compare_map[clazz.devstatus]:
                # First configure the service
                self.configure_service_for_module(clazz.id)
                # Now activate it
                self._services[clazz.id] = clazz(
                    active=clazz.id in self.specter.config.get("services", []),
                    specter=self.specter,
                )
                logger.info(f"Service {clazz.__name__} activated ({clazz.devstatus})")
            else:
                logger.info(
                    f"Service {clazz.__name__} not activated due to devstatus ( {self.devstatus_threshold} > {clazz.devstatus} )"
                )

        # Configure and instantiate the one and only ServiceEncryptedStorageManager
        try:
            ServiceEncryptedStorageManager.configure_instance(
                specter.data_folder, specter.user_manager
            )
        except ConfigurableSingletonException as e:
            # Test suite triggers multiple calls; ignore for now.
            pass
        logger.info("----> finished service discovery <----")

    @classmethod
    def configure_service_for_module(cls, service_id):
        """searches for ConfigClasses in the module-Directory and merges its config in the global config"""
        try:
            module = import_module(
                f"cryptoadvance.specter.services.{service_id}.config"
            )
        except ModuleNotFoundError:
            logger.warning(
                f"Service {service_id} does not have a service Configuration! Skipping!"
            )
            return
        main_config_clazz_name = app.config.get("SPECTER_CONFIGURATION_CLASS_FULLNAME")
        main_config_clazz_slug = main_config_clazz_name.split(".")[-1]
        potential_config_classes = []
        for attribute_name in dir(module):
            attribute = getattr(module, attribute_name)
            if isclass(attribute):
                clazz = attribute
                potential_config_classes.append(clazz)
                if clazz.__name__.endswith(
                    main_config_clazz_slug
                ):  # e.g. BaseConfig or DevelopmentConfig
                    cls.import_config(clazz)
                    return

        logger.warning(
            f"Could not find a configuration for Service {module} ... trying parent-classes of main-config"
        )
        config_module = import_module(".".join(main_config_clazz_name.split(".")[0:-1]))

        config_clazz = getattr(config_module, main_config_clazz_slug)
        config_candidate_class = config_clazz.__bases__[0]
        while config_candidate_class != object:
            for clazz in potential_config_classes:
                if clazz.__name__.endswith(config_candidate_class.__name__):
                    cls.import_config(clazz)
                    return
            config_candidate_class = config_candidate_class.__bases__[0]

    @classmethod
    def import_config(cls, clazz):
        logger.info(f"Loading Service-specific configuration from {clazz}")
        for key in dir(clazz):
            if key.isupper():
                if app.config.get(key):
                    raise Exception(
                        f"Config {clazz} tries to override existing key {key}"
                    )
                app.config[key] = getattr(clazz, key)
                logger.debug(f"setting {key} = {app.config[key]}")

    @property
    def services(self) -> Dict[str, Service]:
        return self._services

    @property
    def services_sorted(self):
        service_names = sorted(
            self._services, key=lambda s: self._services[s].sort_priority
        )
        return [self._services[s] for s in service_names]

    def user_has_encrypted_storage(self, user: User) -> bool:
        """Looks for any data for any service in the User's ServiceEncryptedStorage.
        This check works even if the user doesn't have their plaintext_user_secret
        available."""
        encrypted_data = (
            ServiceEncryptedStorageManager.get_instance().get_raw_encrypted_data(user)
        )
        print(f"encrypted_data: {encrypted_data} for {user}")
        return encrypted_data != {}

    def set_active_services(self, service_names_active):
        logger.debug(f"Setting these services active: {service_names_active}")
        self.specter.update_services(service_names_active)
        for _, service in self.services.items():
            logger.debug(
                f"Setting service '{service.id}' active to {service.id in service_names_active}"
            )
            service.active = service.id in service_names_active

    def get_service(self, service_id: str) -> Service:
        if service_id not in self._services:
            # TODO: better error handling?
            raise Exception(f"No such Service: '{service_id}'")
        return self._services[service_id]

    def remove_all_services_from_user(self, user: User):
        """
        Clears User.services and `user_secret`; wipes the User's
        ServiceEncryptedStorage.
        """
        # Don't show any Services on the sidebar for the admin user
        user.services.clear()

        # Reset as if we never had any encrypted storage
        user.delete_user_secret(autosave=False)
        user.save_info()

        if self.user_has_encrypted_storage(user=user):
            # Encrypted Service data is now orphaned since there is no
            # password. So wipe it from the disk.
            ServiceEncryptedStorageManager.get_instance().delete_all_service_data(user)



import logging
from importlib import import_module
from inspect import isclass
import os
from pathlib import Path
from pkgutil import iter_modules

logger = logging.getLogger(__name__)


def _get_module_from_class(clazz):
    return import_module(clazz.__module__)


def get_package_dir_for_subclasses_of(clazz):
    """There are two occasions where this is used: migrations and service-classes. Depending on the clazz
    this function is returning package_directories where subclasses from clazz are supposed to be located
    * subclasses of SpecterMigration are located in a subpackage called migrations
    * subclasses of Servuce are located in subpackages of package cryptoadvance.specter.services
    I have to admit that this is not a pure util class as it's containing business-logic
    """
    if clazz.__name__ == "SpecterMigration":
        return str(
            Path(
                Path(_get_module_from_class(clazz).__file__).resolve().parent,
                "migrations",
            ).resolve()
        )
    elif clazz.__name__ == "Service":
        return str(
            Path(
                import_module("cryptoadvance.specter.services").__file__
            ).parent.resolve()
        )


def get_subclasses_for_class(clazz):
    """Returns all subclasses of class clazz located in the specific package for that class"""
    class_list = []
    loopdir = Path(__file__).resolve()
    package_dir = get_package_dir_for_subclasses_of(clazz)
    logger.info(f"Collecting subclasses of {clazz.__name__} ...")
    for (_, module_name, _) in iter_modules(
        [package_dir]
    ):  # import the module and iterate through its attributes
        # logger.debug(f"Iterating on {module_name} ")
        if clazz.__name__ == "Service":
            try:

                module = import_module(
                    f"cryptoadvance.specter.services.{module_name}.service"
                )
            except ModuleNotFoundError:
                logger.debug(
                    f"No Service Impl found in cryptoadvance.specter.services.{module_name}. Skipping!"
                )
                continue
        elif clazz.__name__ == "SpecterMigration":
            module = import_module(
                f"cryptoadvance.specter.util.migrations.{module_name}"
            )
        for attribute_name in dir(module):
            attribute = getattr(module, attribute_name)
            if isclass(attribute):
                if (
                    issubclass(attribute, clazz)
                    and not attribute.__name__ == clazz.__name__
                ):
                    class_list.append(attribute)
                    logger.info(f"  Found class {attribute.__name__}")
    return class_list
